import numpy as np
import jax.numpy as jnp
import jax
import io
from numpy.typing import NDArray
from returns.result import Result, Success, Failure, safe
from typing import Optional
from collections import namedtuple
import time
from functools import partial
import chex
import mlflow

from .common import initialise_random_tensor

@chex.dataclass(frozen=True)
class KohonenParams:
    initial_neighbourhood_radius: float
    initial_learning_rate: float
    exp_time_const: float
    width: int
    height: int
    num_iterations: int
    feature_size: int

Node = namedtuple("Node", ['i', 'j'])

def _find_best_matching_unit(weights: NDArray[np.float32], x: NDArray[np.float32], 
                             width: int, height: int) -> Node:
    """Finds the network node that is currently most similar to `x`."""
    bmu = np.argmin(jnp.sum((weights - x) ** 2, axis=2))
    return Node(*jnp.unravel_index(bmu, (height, width)))

@partial(jax.jit, static_argnames=("params",))
def _update_weights(
    x: NDArray[np.float32],
    weights: NDArray[np.float32],
    neighbourhood_radius: float,
    learning_rate: float,
    params: KohonenParams
) -> NDArray[np.float32]:
    """Updates weights given the current input vector, `x`."""
    weights = weights.copy()

    bmu = _find_best_matching_unit(weights, x, params.width, params.height)

    i_indices, j_indices = jnp.indices([params.height, params.width])
    distances = jnp.sqrt((i_indices - bmu.i) ** 2 + (j_indices - bmu.j) ** 2)
    influences = jnp.exp(-distances ** 2 / (2 * (neighbourhood_radius ** 2)))
    weights += learning_rate * influences[..., jnp.newaxis] * (x - weights)

    return weights

def _calculate_neighbourhood_radius(iteration: int, params: KohonenParams):
    """Uses exponential decay to compute radius about best matching unit within
    which nodes will be considered neighbours for a given iteration."""
    return (params.initial_neighbourhood_radius * 
            np.exp(-iteration / params.exp_time_const))

def _calculate_learning_rate(iteration: int, params: KohonenParams):
    """Employes exponential decay to calculate learning rate for a given iteration."""
    return (params.initial_learning_rate * 
            np.exp(-iteration / params.exp_time_const))

@safe
def train_kohonen_network(
    X: NDArray[np.float32], 
    params: KohonenParams,
    use_mlflow: bool = False
) -> NDArray[np.float32]:
    """Trains a Kohonen Network (self-organising map) and logs to mlflow.

    Args:
        X: An (m, n) matrix of training samples where m is the number of samples
            and n is the feature vector size.
        params: A KohonenTrainingParams object containing parameters
            for model training.
        use_mlflow: Toggle MLFlow logging.

    Returns:
        A (height, width, n) tensor representing the network of learned weights,
        or an exception.
    """
    if use_mlflow:
        mlflow.log_params({
            "initial_neighbourhood_radius": params.initial_neighbourhood_radius,
            "initial_learning_rate": params.initial_learning_rate,
            "exp_time_const": params.exp_time_const,
            "width": params.width,
            "height": params.height,
            "num_iterations": params.num_iterations,
            "feature_size": params.feature_size
        })

        start_time = time.perf_counter()

    weights = initialise_random_tensor(params.height, params.width, params.feature_size)

    for t in range(params.num_iterations):
        neighbourhood_radius = _calculate_neighbourhood_radius(t, params)
        learning_rate = _calculate_learning_rate(t, params)

        if use_mlflow:
            mlflow.log_metrics({
                "neighbourhood_radius": neighbourhood_radius,
                "learning_rate": learning_rate
            }, step=t)
        
        for x in X:
            # For current iteration parameters, updates network weights using current sample x
            weights = _update_weights(x, weights, neighbourhood_radius, learning_rate, params)
    
    if use_mlflow:
        execution_time = time.perf_counter() - start_time
        mlflow.log_metric("execution_time", execution_time)
        
    # Convert from JAX array to NumPy array
    return np.array(weights)

def create_kohonen_params(
    X: NDArray[np.float32],
    width: int,
    height: int,
    num_iterations: int,
    initial_learning_rate: float = 0.1
) -> Result[KohonenParams, ValueError]:
    """Creates safe and valid KohonenParams using desired hyperparameters or returns 
    an error is given hyperparameters are not conducive."""
    if width <= 2:
        return Failure(ValueError("Width must be greater than two."))
    if height <= 2:
        return Failure(ValueError("Height must be greater than two."))
    if num_iterations <= 0:
        return Failure(ValueError("Number of iterations must be greater than zero."))
    if initial_learning_rate <= 0:
        return Failure(ValueError("Initial learning rate must be greater than zero."))
    if len(X.shape) != 2:
        return Failure(ValueError("Can only create KohonenParams for two dimensional input data."))

    initial_neighbourhood_radius = max(width, height) / 2
    return Success(KohonenParams(
        initial_neighbourhood_radius=initial_neighbourhood_radius,
        initial_learning_rate=initial_learning_rate,
        exp_time_const=num_iterations / np.log(initial_neighbourhood_radius),
        width=width,
        height=height,
        num_iterations=num_iterations,
        feature_size=X.shape[-1]
    ))